#include <cmath>

using namespace std;

#include "expr.hh"
#include "basfun.hh"

expr::expr()
{
}
expr::~expr()
{
}

long double   expr_number::get_f() const { return type==t_float ? float_val : int_val; }
long long int expr_number::get_i() const { return type==t_int   ? int_val : BasCINT(float_val); }

exprp expr_binary::optimized()
{
    expr_string* s0 = dynamic_cast<expr_string*> (&*p0);
    expr_string* s1 = dynamic_cast<expr_string*> (&*p1);
    switch(token)
    {
        #define stringop(t,op,p) if(s0&&s1) return new t((s0->get() op s1->get()) p)
        case '+':            stringop(expr_string,+,);
        case '<':            stringop(expr_number,<,?-1:0);
        case '>':            stringop(expr_number,>,?-1:0);
        case '=':            stringop(expr_number,==,?-1:0);
        case LESSOREQUAL:    stringop(expr_number,<=,?-1:0);
        case GREATEROREQUAL: stringop(expr_number,>=,?-1:0);
        case NOTEQUAL:       stringop(expr_number,!=,?-1:0);
        #undef stringop
        default: break;
    }
    
    expr_number* n0 = dynamic_cast<expr_number*> (&*p0);
    expr_number* n1 = dynamic_cast<expr_number*> (&*p1);
    switch(token)
    {
        #define numberop(t,op,p) if(n0&&n1) return new t((n0->get_f() op n1->get_f()) p)
        case '+': numberop(expr_number,+,);
        case '-': numberop(expr_number,-,);
        case '*': numberop(expr_number,*,);
        case '/': numberop(expr_number,/,);

        case '<':            numberop(expr_number,<,?-1:0);
        case '>':            numberop(expr_number,>,?-1:0);
        case '=':            numberop(expr_number,==,?-1:0);
        case LESSOREQUAL:    numberop(expr_number,<=,?-1:0);
        case GREATEROREQUAL: numberop(expr_number,>=,?-1:0);
        case NOTEQUAL:       numberop(expr_number,!=,?-1:0);
        
        case MOD: if(n0&&n1) return new expr_number(BasMODop(n0->get_i(), n1->get_i()));
        
        case XOR: if(n0&&n1) return new expr_number(n0->get_i() ^ n1->get_i());
        case OR:  if(n0&&n1) return new expr_number(n0->get_i() | n1->get_i());
        case AND: if(n0&&n1) return new expr_number(n0->get_i() & n1->get_i());
        
        case '\\': if(n0&&n1) return new expr_number(BasINT(n0->get_i() / n1->get_i()));
        
        default: break;
    }
    return expr::optimized();
}

exprp expr_builtin_numeric::optimized()
{
    expr_number* n0 =
        params.size() >= 1
        ? dynamic_cast<expr_number*> (&*params[0])
        : 0;
    expr_string* s0 =
        params.size() >= 1
        ? dynamic_cast<expr_string*> (&*params[0])
        : 0;
    
    switch(token)
    {
        case ABS: if(n0) return new expr_number(fabsl(n0->get_f()));
        case ATN: if(n0) return new expr_number(atanl(n0->get_f()));
        case CINT: if(n0) return new expr_number(n0->get_i());
        case CLNG: if(n0) return new expr_number(n0->get_i());
        case CSNG: if(n0) return new expr_number(n0->get_f());
        case CDBL: if(n0) return new expr_number(n0->get_f());
        case COS: if(n0) return new expr_number(cosl(n0->get_f()));
        case EXP: if(n0) return new expr_number(expl(n0->get_f()));
        case FIX: if(n0) return new expr_number(truncl(n0->get_f()));
        case LOG: if(n0) return new expr_number(logl(n0->get_f()));
        case SQR: if(n0) return new expr_number(sqrtl(n0->get_f()));
        case TAN: if(n0) return new expr_number(tanl(n0->get_f()));
        case VAL: if(s0) return new expr_number(BasVAL(s0->get()));
        default: break;
    }
    return expr::optimized();
}

exprp expr_builtin_string::optimized()
{
    expr_number* n0 =
        params.size() >= 1
        ? dynamic_cast<expr_number*> (&*params[0])
        : 0;

    expr_number* n1 =
        params.size() >= 2
        ? dynamic_cast<expr_number*> (&*params[1])
        : 0;

    expr_string* s0 =
        params.size() >= 1
        ? dynamic_cast<expr_string*> (&*params[0])
        : 0;

#if 0

        /* Pure functions: functions that only depend on their parameter;
         * can be calculated at compile-time. */
        return (token == LTRIM_S
             || token == RTRIM_S
             || token == LCASE_S
             || token == UCASE_S
             || token == INSTR
             || token == MID_S
             || token == RIGHT_S
             || token == STRING_S
             || token == CHR_S
             || token == HEX_S
             || token == MKD_S
             || token == MKI_S
             || token == MKL_S
             || token == MKS_S
             || token == OCT_S
             || token == SPACE_S
             || token == STR_S
               );
#endif
    switch(token)
    {
        case LTRIM_S: if(s0) return new expr_string(BasLTRIM(s0->get()));
        case RTRIM_S: if(s0) return new expr_string(BasRTRIM(s0->get()));
        case LCASE_S: if(s0) return new expr_string(BasLCASE(s0->get()));
        case UCASE_S: if(s0) return new expr_string(BasUCASE(s0->get()));
        
        case STR_S: if(n0) return new expr_string(BasSTRs(n0->get_f()));
        case CHR_S: if(n0) return new expr_string(BasCHRs(n0->get_i()));
        
        default: break;
    }
    return expr::optimized();
}
